import logging
from typing import override

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.models.audit_log import AuditLogRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.reporters.base import (
    AbstractReporter,
    FinishedActionMessage,
    StartedActionMessage,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AuditLogReporter(AbstractReporter):
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def _generate_log(self, action_message: FinishedActionMessage) -> None:
        async with self._db.begin_session() as db_sess:
            db_row = AuditLogRow(
                action_id=action_message.action_id,
                entity_type=action_message.entity_type,
                operation=action_message.operation_type,
                created_at=action_message.created_at,
                entity_id=action_message.entity_id,
                request_id=action_message.request_id,
                description=action_message.description,
                status=action_message.status,
                duration=action_message.duration,
            )

            db_sess.add(db_row)
            await db_sess.flush()

    @override
    async def report_started(self, message: StartedActionMessage) -> None:
        pass

    @override
    async def report_finished(self, message: FinishedActionMessage) -> None:
        await self._generate_log(message)



from __future__ import annotations

import enum
import logging
import uuid
from datetime import datetime, timedelta
from typing import Optional

import sqlalchemy as sa

from ai.backend.logging import BraceStyleAdapter

from .base import (
    GUID,
    Base,
    IDColumn,
    StrEnumType,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("AuditLogRow",)


class AuditLogEntityType(enum.StrEnum):
    IMAGE = "image"
    CONTAINER_REGISTRY = "container_registry"
    DOMAIN = "domain"
    GROUP = "group"
    AGENT = "agent"
    KEYPAIR_RESOURCE_POLICY = "keypair_resource_policy"
    PROJECT_RESOURCE_POLICY = "project_resource_policy"
    USER_RESOURCE_POLICY = "user_resource_policy"
    RESOURCE_PRESET = "resource_preset"
    SESSION = "session"
    USER = "user"
    VFOLDER = "vfolder"
    VFOLDER_INVITATION = "vfolder_invitation"


class OperationStatus(enum.StrEnum):
    SUCCESS = "success"
    ERROR = "error"
    UNKNOWN = "unknown"
    RUNNING = "running"


class AuditLogRow(Base):
    __tablename__ = "audit_logs"

    id = IDColumn("id")

    entity_type = sa.Column("entity_type", sa.String, index=True, nullable=False)
    operation = sa.Column("operation", sa.String, index=True, nullable=False)

    entity_id = sa.Column(
        "entity_id",
        sa.String,
        nullable=True,
        index=True,
    )

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
        index=True,
    )

    action_id = sa.Column("action_id", GUID, nullable=False)
    request_id = sa.Column("request_id", sa.String, nullable=True)
    description = sa.Column("description", sa.String, nullable=False)
    duration = sa.Column("duration", sa.Interval, nullable=True)

    status = sa.Column(
        "status",
        StrEnumType(OperationStatus),
        nullable=False,
    )

    def __init__(
        self,
        entity_type: str,
        operation: str,
        action_id: uuid.UUID,
        description: str,
        created_at: datetime,
        status: OperationStatus,
        entity_id: Optional[str | uuid.UUID] = None,
        request_id: Optional[str] = None,
        duration: Optional[timedelta] = None,
    ):
        self.entity_type = entity_type
        self.operation = operation
        self.action_id = action_id
        self.entity_id = str(entity_id) if isinstance(entity_id, uuid.UUID) else entity_id
        self.request_id = request_id
        self.description = description
        self.duration = duration
        self.status = status
        self.created_at = created_at

    def __str__(self) -> str:
        return (
            f"AuditLogRow("
            f"entity_type: {self.entity_type}, "
            f"operation: {self.operation}, "
            f"created_at: {self.created_at}, "
            f"entity_id: {self.entity_id}, "
            f"action_id: {self.action_id}, "
            f"request_id: {self.request_id}, "
            f"description: {self.description}, "
            f"duration: {self.duration}, "
            f"status: {self.status.value}"
            f")"
        )

    def __repr__(self) -> str:
        return self.__str__()



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.common.types import AgentId
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.models.agent import AgentRow
from ai.backend.manager.services.agent.actions.base import AgentAction


@dataclass
class SyncAgentRegistryAction(AgentAction):
    agent_id: AgentId

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        return "sync_agent_registry"


@dataclass
class SyncAgentRegistryActionResult(BaseActionResult):
    # TODO: Add proper type
    result: Any
    agent_row: AgentRow

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.agent_row.id)



import logging
import smtplib
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from email.mime.text import MIMEText
from typing import Final, override

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.models.audit_log import OperationStatus
from ai.backend.manager.reporters.base import (
    AbstractReporter,
    FinishedActionMessage,
    StartedActionMessage,
)
from ai.backend.manager.types import SMTPTriggerPolicy

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


_UNDEFINED_VALUE: Final[str] = "(undefined)"


@dataclass
class SMTPSenderArgs:
    host: str
    port: int
    username: str
    password: str
    sender: str
    recipients: list[str]
    use_tls: bool
    template: str
    max_workers: int


class SMTPSender:
    def __init__(self, args: SMTPSenderArgs) -> None:
        self._config = args
        self._executor = ThreadPoolExecutor(max_workers=self._config.max_workers)

    def send_email(self, subject: str, email_body: str) -> None:
        self._executor.submit(self._send_email, subject, email_body)

    def _send_email(self, subject: str, email_body: str) -> None:
        message = MIMEText(email_body, "plain", "utf-8")
        message["Subject"] = subject
        message["From"] = self._config.sender
        message["To"] = ",".join(self._config.recipients)

        try:
            with smtplib.SMTP(self._config.host, self._config.port) as server:
                if self._config.use_tls:
                    server.starttls()
                server.login(self._config.username, self._config.password)
                server.send_message(
                    message,
                    from_addr=self._config.sender,
                    to_addrs=self._config.recipients,
                )
        except Exception as e:
            log.error(f"Failed to send email: {e}")


class SMTPReporter(AbstractReporter):
    _mail_template: str
    _smtp_sender: SMTPSender

    def __init__(self, args: SMTPSenderArgs, trigger_policy: SMTPTriggerPolicy) -> None:
        self._smtp_sender = SMTPSender(args)
        self._trigger_policy = trigger_policy
        self._mail_template = args.template

    def _create_body_from_template(self, message: FinishedActionMessage) -> str:
        template = self._mail_template

        template = template.replace("{{ action_id }}", str(message.action_id))
        template = template.replace("{{ action_type }}", message.action_type)
        template = template.replace(
            "{{ entity_id }}", str(message.entity_id) if message.entity_id else _UNDEFINED_VALUE
        )
        template = template.replace("{{ request_id }}", message.request_id or _UNDEFINED_VALUE)
        template = template.replace("{{ entity_type }}", message.entity_type)
        template = template.replace("{{ operation_type }}", message.operation_type)
        template = template.replace("{{ created_at }}", str(message.created_at))
        template = template.replace("{{ ended_at }}", str(message.ended_at))
        template = template.replace("{{ duration }}", str(message.duration))
        template = template.replace("{{ status }}", message.status.value)
        template = template.replace("{{ description }}", message.description)

        return template

    def _make_subject(self, action_type: str) -> str:
        return f"Backend.AI SMTP Log Alert ({action_type})"

    @override
    async def report_started(self, message: StartedActionMessage) -> None:
        pass

    @override
    async def report_finished(self, message: FinishedActionMessage) -> None:
        if self._trigger_policy == SMTPTriggerPolicy.ON_ERROR:
            if message.status == OperationStatus.ERROR:
                subject = self._make_subject(message.action_type)
                body = (
                    "Action has resulted in an error.\n\n"
                    f"{self._create_body_from_template(message)}"
                )
                self._smtp_sender.send_email(subject, body)
                return

        subject = self._make_subject(message.action_type)
        body = f"Action has finished.\n\n{self._create_body_from_template(message)}"
        self._smtp_sender.send_email(subject, body)
